import click
import os

from stactools.cop_dem import stac
from stactools.cop_dem import constants as co


def create_cop_dem_command(cli):
    """Creates the COP DEM stactools commands."""

    @cli.group("cop-dem", short_help="Work with Copernicus DEM data.")
    def cop_dem():
        pass

    @cop_dem.command(
        "create-item",
        short_help="Creates a STAC item from an Copernicus DEM tile.")
    @click.argument("source")
    @click.argument("destination")
    @click.option("--validate/--no-validate",
                  default=True,
                  help="Validate the item before saving")
    @click.option("--host",
                  type=click.Choice(co.COP_DEM_HOST.keys(),
                                    case_sensitive=False),
                  default=None,
                  help="Set PROVIDER HOST")
    def create_item_command(source: str, destination: str, validate: bool,
                            host: str):
        """ Creates a STAC Item

        source: Path to input item

        destination: File to output json

        """
        item = stac.create_item(source, host=host)
        if validate:
            item.validate()
        item.save_object(dest_href=destination)

    @cop_dem.command("create-collection",
                     short_help="Creates a STAC Collection for Copernicus DEM."
                     )
    @click.argument("product", type=click.Choice(['glo-30', 'glo-90']))
    @click.argument("destination", type=click.Path(exists=True))
    @click.option("-u",
                  "--url",
                  default='',
                  type=str,
                  help="Root url to prepend to all records")
    @click.option("--validate/--no-validate",
                  default=True,
                  help="Validate the item before saving")
    @click.option("--host",
                  type=click.Choice(co.COP_DEM_HOST.keys(),
                                    case_sensitive=False),
                  default=None,
                  help="Set PROVIDER HOST")
    def create_collection_command(destination: str, product: str, url: str,
                                  validate: bool, host: str):
        """ Creates a STAC Collection

        product: The DEM product, glo-30 or glo-90

        destination: Folder to output json

        """
        collection = stac.create_collection(product, host=host)
        json_path = os.path.join(destination, f'{collection.id}.json')
        collection.set_self_href(
            os.path.join(url, collection.id, os.path.basename(json_path)))
        if validate:
            collection.validate()
        collection.save_object(dest_href=json_path)

    return cop_dem



from datetime import datetime
from typing import Optional

from pystac import Provider, Link, ProviderRole
from pystac.extensions.item_assets import AssetDefinition
from pystac.utils import str_to_datetime
from pystac.media_type import MediaType

COP_DEM_PRODUCTS = ['glo-30', 'glo-90']

COP_DEM_SPATIAL_EXTENT = [[-180., -90., 180., 90.]]

COP_DEM_COLLECTION_START: Optional[datetime] = str_to_datetime(
    "2021-04-22T00:00:00Z")

COP_DEM_COLLECTION_END: Optional[datetime] = str_to_datetime(
    "2021-04-22T00:00:00Z")

COP_DEM_TEMPORAL_EXTENT = [COP_DEM_COLLECTION_START,
                           COP_DEM_COLLECTION_END]  # TODO: find the dates

COP_DEM_PLATFORM = "tandem-x"

COP_DEM_EPSG = 4326

COP_DEM_KEYWORDS = ['DEM', 'COPERNICUS', 'DSM', 'Elevation']

COP_DEM_PROVIDERS = [
    Provider("European Space Agency",
             roles=[ProviderRole.LICENSOR],
             url=("https://spacedata.copernicus.eu/documents/20126/0/"
                  "CSCDA_ESA_Mission-specific+Annex.pdf")),
    Provider("Sinergise",
             roles=[ProviderRole.PRODUCER, ProviderRole.PROCESSOR],
             url="https://registry.opendata.aws/copernicus-dem/"),
]

# Various HOST options
COP_DEM_HOST = {
    "OT":
    Provider("OpenTopography",
             roles=[ProviderRole.HOST],
             url=("https://portal.opentopography.org/"
                  "datasetMetadata?otCollectionID=OT.032021.4326.1")),
    "AWS":
    Provider("Sinergise",
             roles=[ProviderRole.HOST],
             url="https://registry.opendata.aws/copernicus-dem/")
}

COP_DEM_LINKS = [
    Link(
        "handbook",
        "https://object.cloud.sdsc.edu/v1/AUTH_opentopography/www/metadata/Copernicus_metadata.pdf",
        "application/pdf",
        "Copernicus DEM User handbook",
        extra_fields={"description": "Also includes data usage information"}),
    Link(
        "license",
        "https://spacedata.copernicus.eu/documents/20123/121286/CSCDA_ESA_Mission-specific+Annex_31_Oct_22.pdf",  # noqa: E501
        "Copernicus Data Access")
]

COP_DEM_ASSETS = {
    "data": AssetDefinition({
        "type": MediaType.COG,
        "roles": ["data"],
    })
}

COP_DEM_DESCRIPTION = '''The Copernicus DEM is a Digital Surface Model (DSM) which represents the surface of the Earth including buildings, infrastructure and vegetation. We provide two instances of Copernicus DEM named GLO-30 Public and GLO-90. GLO-90 provides worldwide coverage at 90 meters. GLO-30 Public provides limited worldwide coverage at 30 meters because a small subset of tiles covering specific countries are not yet released to the public by the Copernicus Programme. Note that in both cases ocean areas do not have tiles, there one can assume height values equal to zero. Data is provided as Cloud Optimized GeoTIFFs and comes from Copernicus DEM 2021 release.'''  # noqa: E501



import os.path
from tempfile import TemporaryDirectory

import pystac
from stactools.testing import CliTestCase

from stactools.cop_dem import commands
from tests import test_data


class CreateItemTest(CliTestCase):

    def create_subcommand_functions(self):
        return [commands.create_cop_dem_command]

    def test_create_item(self):
        infile = test_data.get_external_data(
            "Copernicus_DSM_COG_10_N53_00_W115_00_DEM.tif")
        with TemporaryDirectory() as temporary_directory:
            outfile = os.path.join(temporary_directory, "item.json")
            args = ["cop-dem", "create-item", infile, outfile]
            result = self.run_command(args)
            self.assertEqual(result.exit_code, 0)
            item = pystac.read_file(outfile)
            item.validate()

    def test_create_collection(self):
        with TemporaryDirectory() as temporary_directory:
            outfile = os.path.join(temporary_directory, "cop-dem-glo-30.json")
            print(outfile)
            args = [
                "cop-dem", "create-collection", "glo-30", temporary_directory
            ]
            result = self.run_command(args)
            self.assertEqual(result.exit_code, 0)
            collection = pystac.read_file(outfile)
            collection.validate()



import os.path
import re
from typing import Optional

from pystac import (
    CatalogType,
    Collection,
    Extent,
    Asset,
    Summaries,
    SpatialExtent,
    TemporalExtent,
)
from pystac.extensions.grid import GridExtension
from pystac.extensions.projection import ProjectionExtension
from pystac.extensions.item_assets import ItemAssetsExtension
from pystac.extensions.raster import (
    DataType,
    RasterBand,
    RasterExtension,
    Sampling,
)
from pystac.media_type import MediaType

import rasterio
from shapely.geometry import mapping, box, shape as make_shape
from pystac import Item

from stactools.core.io import ReadHrefModifier
from stactools.cop_dem import constants as co


def create_item(
    href: str,
    read_href_modifier: Optional[ReadHrefModifier] = None,
    host: Optional[str] = None,
) -> Item:
    """Creates a STAC Item from a single tile of Copernicus DEM data."""
    if read_href_modifier:
        modified_href = read_href_modifier(href)
    else:
        modified_href = href
    with rasterio.open(modified_href) as dataset:
        if dataset.crs.to_epsg() != co.COP_DEM_EPSG:
            raise ValueError(f"Dataset {href} is not EPSG:{co.COP_DEM_EPSG}, "
                             "which is required for Copernicus DEM data")
        bbox = list(dataset.bounds)
        geometry = mapping(box(*bbox))
        transform = dataset.transform
        shape = dataset.shape
        item = Item(
            id=os.path.splitext(os.path.basename(href))[0],
            geometry=geometry,
            bbox=bbox,
            datetime=co.COP_DEM_COLLECTION_START,
            properties={},
            stac_extensions=[],
        )

    # resolution in arc seconds (not meters!), which is and 30 for GLO-90 and 10 for GLO-30
    p = re.compile(
        r"Copernicus_DSM_COG_(?P<res>\d{2})_(?P<northing>[NS]\d{2})_00_(?P<easting>[EW]\d{3})_00_DEM.*"  # noqa: E501
    )
    if m := p.match(os.path.basename(href)):
        res = m.group("res")
        if res == "30":
            gsd = 90
        elif res == "10":
            gsd = 30
        else:
            raise ValueError(f"unknown resolution {res}")

        northing = m.group("northing")
        easting = m.group("easting")
    else:
        raise ValueError(f"unable to parse {href}")

    item.add_links(co.COP_DEM_LINKS)
    item.common_metadata.platform = co.COP_DEM_PLATFORM
    item.common_metadata.gsd = gsd

    if host and host not in co.COP_DEM_HOST.keys():
        raise ValueError(
            f"Invalid host: {host}. Must be one of {list(co.COP_DEM_HOST.keys())}"
        )
    if host and (host_provider := co.COP_DEM_HOST.get(host)):
        providers = [*co.COP_DEM_PROVIDERS, host_provider]
    else:
        providers = co.COP_DEM_PROVIDERS

    item.common_metadata.providers = providers
    item.common_metadata.license = "proprietary"

    data_asset = Asset(
        href=href,
        title="Data",
        description=None,
        media_type=MediaType.COG,
        roles=["data"],
    )
    data_bands = RasterBand.create(
        sampling=Sampling.POINT,
        data_type=DataType.FLOAT32,
        spatial_resolution=gsd,
        unit="meter",
    )

    item.add_asset("data", data_asset)
    RasterExtension.ext(data_asset, add_if_missing=True).bands = [data_bands]

    projection = ProjectionExtension.ext(item, add_if_missing=True)
    projection.epsg = co.COP_DEM_EPSG
    projection.transform = transform[0:6]
    projection.shape = shape

    centroid = make_shape(item.geometry).centroid
    projection.centroid = {
        "lat": round(centroid.y, 1),
        "lon": round(centroid.x, 1)
    }

    grid = GridExtension.ext(item, add_if_missing=True)
    grid.code = f"CDEM-{northing}{easting}"

    return item


def create_collection(product: str, host: Optional[str] = None) -> Collection:
    """Create a STAC Collection

    This function includes logic to extract all relevant metadata from
    an asset describing the STAC collection and/or metadata coded into an
    accompanying constants.py file.

    See `Collection<https://pystac.readthedocs.io/en/latest/api.html#collection>`_.

    Args:
        Product (str): glo-30 or glo-90

    Returns:
        Item: STAC Item object

    Returns:
        Collection: STAC Collection object
    """
    if product.lower() == "glo-30":
        summaries = {
            "gsd": [30],
            "platform": [co.COP_DEM_PLATFORM],
            # "instruments": ,
        }
    elif product.lower() == "glo-90":
        summaries = {
            "gsd": [90],
            "platform": [co.COP_DEM_PLATFORM],
            # "instruments": ,
        }
    else:
        raise ValueError(
            f"{product} is not a valid product. Must be one of {co.COP_DEM_PRODUCTS}"
        )

    # Allow host to be selected by cli option
    if host and host not in co.COP_DEM_HOST.keys():
        raise ValueError(
            f"Invalid host: {host}. Must be one of {list(co.COP_DEM_HOST.keys())}"
        )
    if host and (host_provider := co.COP_DEM_HOST.get(host)):
        providers = [*co.COP_DEM_PROVIDERS, host_provider]
    else:
        providers = co.COP_DEM_PROVIDERS

    collection = Collection(
        id=f"cop-dem-{product.lower()}",
        title=f"Copernicus DEM {product.upper()}",
        description=co.COP_DEM_DESCRIPTION,
        license="proprietary",
        keywords=co.COP_DEM_KEYWORDS,
        catalog_type=CatalogType.RELATIVE_PUBLISHED,
        summaries=Summaries(summaries),
        extent=Extent(
            SpatialExtent(co.COP_DEM_SPATIAL_EXTENT),
            TemporalExtent([co.COP_DEM_TEMPORAL_EXTENT]),
        ),
        providers=providers,
        stac_extensions=[
            ItemAssetsExtension.get_schema_uri(),
            ProjectionExtension.get_schema_uri(),
            RasterExtension.get_schema_uri(),
        ],
    )

    collection.add_links(co.COP_DEM_LINKS)

    assets = ItemAssetsExtension.ext(collection, add_if_missing=True)
    assets.item_assets = co.COP_DEM_ASSETS

    return collection



import datetime
from unittest import TestCase

from pystac import Provider, MediaType
from pystac.extensions.grid import GridExtension
from pystac.extensions.projection import ProjectionExtension
from pystac.extensions.raster import RasterExtension
from pystac.provider import ProviderRole

from stactools.cop_dem import stac

from tests import test_data


class StacTest(TestCase):

    def setUp(self):
        self.glo30_path = test_data.get_external_data(
            "Copernicus_DSM_COG_10_N53_00_W115_00_DEM.tif")
        self.glo90_path = test_data.get_external_data(
            "Copernicus_DSM_COG_30_N53_00_W115_00_DEM.tif")

    def test_create_glo30_item(self):
        item = stac.create_item(self.glo30_path)
        self.assertEqual(item.id, "Copernicus_DSM_COG_10_N53_00_W115_00_DEM")
        self.assertIsNotNone(item.geometry)
        self.assertEqual(list(item.bbox), [
            -115.00020833333333, 53.00013888888889, -114.00020833333333,
            54.00013888888889
        ])
        self.assertEqual(
            item.datetime,
            datetime.datetime(2021, 4, 22, tzinfo=datetime.timezone.utc))

        common_metadata = item.common_metadata
        self.assertEqual(common_metadata.platform, "tandem-x")
        self.assertEqual(common_metadata.gsd, 30)
        expected_providers = [
            Provider("European Space Agency",
                     roles=[ProviderRole.LICENSOR],
                     url=("https://spacedata.copernicus.eu/documents/20126/0/"
                          "CSCDA_ESA_Mission-specific+Annex.pdf")),
            Provider("Sinergise",
                     roles=[ProviderRole.PRODUCER, ProviderRole.PROCESSOR],
                     url="https://registry.opendata.aws/copernicus-dem/"),
            Provider("OpenTopography",
                     roles=[ProviderRole.HOST],
                     url=("https://portal.opentopography.org/"
                          "datasetMetadata?otCollectionID=OT.032021.4326.1"))
        ]
        for expected, actual in zip(expected_providers,
                                    common_metadata.providers):
            self.assertDictEqual(expected.to_dict(), actual.to_dict())
        self.assertEqual(common_metadata.license, "proprietary")

        projection = ProjectionExtension.ext(item)
        self.assertEqual(projection.epsg, 4326)
        self.assertEqual(projection.shape, (3600, 2400))
        self.assertEqual(list(projection.transform), [
            0.00041666666666666664, 0.0, -115.00020833333333, 0.0,
            -0.0002777777777777778, 54.00013888888889
        ])

        grid = GridExtension.ext(item)
        self.assertEqual(grid.code, "CDEM-N53W115")

        handbook = item.get_single_link("handbook")
        self.assertIsNotNone(handbook)
        self.assertEqual(handbook.title, "Copernicus DEM User handbook")
        self.assertEqual(handbook.rel, "handbook")
        self.assertEqual(
            handbook.href,
            "https://object.cloud.sdsc.edu/v1/AUTH_opentopography/www/metadata"
            "/Copernicus_metadata.pdf")
        self.assertEqual(handbook.media_type, "application/pdf")

        data = item.assets["data"]
        self.assertEqual(data.href, self.glo30_path)
        self.assertEqual(data.title, "Data")
        self.assertIsNone(data.description)
        self.assertEqual(data.media_type, MediaType.COG)
        self.assertEqual(data.roles, ["data"])

        self.assertTrue(GridExtension.has_extension(item))
        self.assertTrue(ProjectionExtension.has_extension(item))
        self.assertTrue(RasterExtension.has_extension(item))

        item.validate()  # raises STACValidationError if not

    def test_create_glo90_item(self):
        item = stac.create_item(self.glo90_path)
        self.assertEqual(item.id, "Copernicus_DSM_COG_30_N53_00_W115_00_DEM")
        self.assertIsNotNone(item.geometry)
        self.assertEqual(
            list(item.bbox),
            [-115.000625, 53.000416666666666, -114.000625, 54.000416666666666])
        self.assertEqual(
            item.datetime,
            datetime.datetime(2021, 4, 22, tzinfo=datetime.timezone.utc))

        common_metadata = item.common_metadata
        self.assertEqual(common_metadata.platform, "tandem-x")
        self.assertEqual(common_metadata.gsd, 90)
        expected_providers = [
            Provider("European Space Agency",
                     roles=[ProviderRole.LICENSOR],
                     url=("https://spacedata.copernicus.eu/documents/20126/0/"
                          "CSCDA_ESA_Mission-specific+Annex.pdf")),
            Provider("Sinergise",
                     roles=[ProviderRole.PRODUCER, ProviderRole.PROCESSOR],
                     url="https://registry.opendata.aws/copernicus-dem/"),
            Provider("OpenTopography",
                     roles=[ProviderRole.HOST],
                     url=("https://portal.opentopography.org/"
                          "datasetMetadata?otCollectionID=OT.032021.4326.1"))
        ]
        for expected, actual in zip(expected_providers,
                                    common_metadata.providers):
            self.assertDictEqual(expected.to_dict(), actual.to_dict())
        self.assertEqual(common_metadata.license, "proprietary")

        projection = ProjectionExtension.ext(item)
        self.assertEqual(projection.epsg, 4326)
        self.assertEqual(projection.shape, (1200, 800))
        self.assertEqual(list(projection.transform), [
            0.00125, 0.0, -115.000625, 0.0, -0.0008333333333333334,
            54.000416666666666
        ])
        self.assertEqual(round(projection.centroid["lat"], 5), 53.5)
        self.assertEqual(round(projection.centroid["lon"], 5), -114.5)

        grid = GridExtension.ext(item)
        self.assertEqual(grid.code, "CDEM-N53W115")

        handbook = item.get_single_link("handbook")
        self.assertIsNotNone(handbook)
        self.assertEqual(handbook.title, "Copernicus DEM User handbook")
        self.assertEqual(handbook.rel, "handbook")
        self.assertEqual(
            handbook.href,
            "https://object.cloud.sdsc.edu/v1/AUTH_opentopography/www/metadata"
            "/Copernicus_metadata.pdf")
        self.assertEqual(handbook.media_type, "application/pdf")

        data = item.assets["data"]
        self.assertEqual(data.href, self.glo90_path)
        self.assertEqual(data.title, "Data")
        self.assertIsNone(data.description)
        self.assertEqual(data.media_type, MediaType.COG)
        self.assertEqual(data.roles, ["data"])

        self.assertTrue(ProjectionExtension.has_extension(item))
        self.assertTrue(RasterExtension.has_extension(item))
        self.assertTrue(GridExtension.has_extension(item))

        item.validate()  # raises STACValidationError if not

    def test_create_item_with_read_href_modifier(self):
        done = False

        def do_it(href):
            nonlocal done
            done = True
            return href

        _ = stac.create_item(self.glo30_path, read_href_modifier=do_it)
        self.assertTrue(done, "Didn't do it")

    def test_create_collection(self):
        # Write tests for each for the creation of a STAC Collection
        # Create the STAC Collection...
        collection = stac.create_collection("glo-30")
        collection.set_self_href("")

        # Check that it has some required attributes
        self.assertEqual(collection.id, "cop-dem-glo-30")
        # self.assertEqual(collection.other_attr...

        # Validate
        collection.validate()



from stactools.core import io

io.use_fsspec()


def register_plugin(registry):
    from stactools.cop_dem import commands
    registry.register_subcommand(commands.create_cop_dem_command)


__version__ = "0.4.1"
